// mandelbrot_fifo.cpp: Mandelbrot set renderer example in SystemC
// Author: Guantao Liu, CECS
//
// Modifications: (most recent first)
// 09/30/16 RD	separate version mandelbrot_fifo.cpp with sc_fifo channels
// 05/26/16 RD  fixed user-defined channels for multiple senders/receivers
// 05/25/16 RD  added support for SIMD vectorization (VECTORIZE)
// 03/08/16 RD  added support for BLACKWHITE images (from book chapter)
// 03/07/16 RD  added support for PAR=128 and PAR=256 (from version by GL)
// 03/07/16 RD  bug fix: use sc_stop() to quit cleanly (exit() is not MT-safe)
// 03/07/16 RD  bug fix: DUT must wait for all workers (wait(main_continue))
// 03/30/15 GL  SystemC version based on mandelbrot.sc (SpecC example by RD)

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>

#include "systemc.h"


// Definitions ///////////////////////////////////////////////////////////


//#define DEBUG         // print debugging messages

//#define JPEG          // convert saved images to .jpg format
//#define DISPLAY       // display calculated image (using SDL)
//#define BLACKWHITE    // generate black&white images (default: color)

//#define VECTORIZE	// OpenMP SIMD vectorization for Xeon Phi

//#define float_t	float
#define float_t		double
//#define float_t	long double

// pretty location "Zom-B" from:
// http://en.wikipedia.org/wiki/File:Mandelbrot_sequence_new.gif
#define ZOM_B_X         -0.743643887037158704752191506114774L
#define ZOM_B_Y         0.131825904205311970493132056385139L

#ifndef WIDTH           // size of the image in pixels
#define WIDTH           640
#endif
#ifndef HEIGHT
#define HEIGHT          512
#endif
#ifndef MAX_ITER        // depth of calculation
//#define MAX_ITER      1024
//#define MAX_ITER      2048
//#define MAX_ITER      3072
#define MAX_ITER        4096
#endif
#ifndef START_LEFT      // initial position over the Mandelbrot set
//#define START_LEFT    -2.2
#define START_LEFT      (-1.6 + ZOM_B_X)
#endif
#ifndef START_RIGHT
//#define START_RIGHT   1.0
#define START_RIGHT     (1.6 + ZOM_B_X)
#endif
#ifndef START_BOTTOM
//#define START_BOTTOM  -1.0
#define START_BOTTOM    (-1.12 + ZOM_B_Y)
#endif
#ifndef START_TOP
//#define START_TOP     1.0
#define START_TOP       (1.12 + ZOM_B_Y)
#endif
#ifndef TARGET_COORDS_X // target coordinates
#define TARGET_COORDS_X ZOM_B_X
#endif
#ifndef TARGET_COORDS_Y
#define TARGET_COORDS_Y ZOM_B_Y
#endif
#ifndef TARGET_ZOOM     // zoom factor per step towards the target
#define TARGET_ZOOM     0.7
#endif
#ifndef NUM_IMAGES      // number of images produced
#define NUM_IMAGES      20
#endif
#ifndef MAX_IMAGE_FILES // number of image files kept
#define MAX_IMAGE_FILES 0
#endif

#define LINEAR_MOTION   // move in equidistant steps (constant speed)
//#define FASTER_MOTION // move in logarithmic steps (slow, then faster and faster)
//#define SLOWER_MOTION // move in logarithmic steps (fast, then slower and slower)

#ifndef PAR             // number of parallel units instantiated
//#define PAR           1
//#define PAR           2
//#define PAR           4
#define PAR             8
//#define PAR           16
//#define PAR           32
//#define PAR           64
//#define PAR           128
//#define PAR           256
#endif

#define USE_SC_THREAD    // SC_THREAD vs. SC_METHOD


#ifdef DISPLAY
#include "sdlsysc.h"
#endif /* DISPLAY */

#ifdef BLACKWHITE
#define NCOLORS         2
#else
#define NCOLORS         16
#endif


// Data structures ///////////////////////////////////////////////////////


#ifdef BLACKWHITE
const unsigned char palette[NCOLORS][3] = {
    //  r    g    b
    {   0,   0,   0 },  // 0, black
    { 255, 255, 255 }}; // 1, white
#else
const unsigned char palette[NCOLORS][3] = {
    //  r    g    b
    {   0,   0,   0 },  // 0, black
    { 127,   0,   0 },  // 1, brown
    { 255,   0,   0 },  // 2, red
    { 255, 127,   0 },  // 3, orange
    { 255, 255,   0 },  // 4, yellow
    { 127, 255,   0 },  // 5, light green
    {   0, 255,   0 },  // 6, green
    {   0, 255, 127 },  // 7, blue green
    {   0, 255, 255 },  // 8, turquoise
    { 127, 255, 255 },  // 9, light blue
    { 255, 255, 255 },  // 10, white
    { 255, 127, 255 },  // 11, pink
    { 255,   0, 255 },  // 12, light pink
    { 127,   0, 255 },  // 13, purple
    {   0,   0, 255 },  // 14, blue
    {   0,   0, 127 }}; // 15, dark blue
#endif

typedef struct Image
{
    unsigned char R[HEIGHT][WIDTH];
    unsigned char G[HEIGHT][WIDTH];
    unsigned char B[HEIGHT][WIDTH];

    Image(void)
    {
        for (int i=0; i<HEIGHT; i++)
            for (int j=0; j<WIDTH; j++)
            {
                R[i][j] = 0;
                G[i][j] = 0;
                B[i][j] = 0;
            }
    }

    Image& operator=(const Image& copy)
    {
        for (int i=0; i<HEIGHT; i++)
            for (int j=0; j<WIDTH; j++)
            {
                R[i][j] = copy.R[i][j];
                G[i][j] = copy.G[i][j];
                B[i][j] = copy.B[i][j];
            }
        return *this;
    }

    friend ostream& operator<<(ostream& os, const Image& im);
} IMAGE;

ostream& operator<<(ostream& os, const Image& im)
{
    for (int i=0; i<HEIGHT; i++)
        for (int j=0; j<WIDTH; j++)
        {
            os << im.R[i][j] << ' ';
            os << im.G[i][j] << ' ';
            os << im.B[i][j] << '\n';
        }
    return os;
}

typedef struct Coordinates
{
    float_t l;  // left bound, e.g. -2.5
    float_t r;  // right bound, e.g. 1.0
    float_t b;  // bottom bound, e.g. -1.0
    float_t t;  // top bound, e.g. 1.0

    friend ostream& operator<<(ostream& os, const Coordinates& co);
} COORDS;

ostream& operator<<(ostream& os, const Coordinates& co)
{
    os << co.l << ' ' << co.r << ' ' << co.b << ' ' << co.t << '\n';
    return os;
}


// Crunch functions //////////////////////////////////////////////////////


// the following Mandelbrot core functions are used below for the cases
// where DISPLAY is turned off (that is, for performance evaluation!);
// if VECTORIZE is enabled, these are tuned to exploit SIMD parallelism
// (05/25/16, RD)

#ifdef VECTORIZE
#pragma omp declare simd
#endif
unsigned int mandel_iter(float_t x0, float_t y0)
{
    float_t x, y;
    float_t xtemp;
    unsigned int iteration;
    const unsigned int max_iteration = MAX_ITER;

    x = 0;
    y = 0;
    iteration = 0;
    while ( x*x + y*y < 2*2  &&  iteration < max_iteration )
    {
        xtemp = x*x - y*y + x0;
        y = 2*x*y + y0;
        x = xtemp;
        iteration = iteration + 1;
    }
    return iteration;
}

void mandel_row(IMAGE *img, COORDS coords, int row)
{
    int col;
    float_t x0, y0;
    unsigned int iteration;
    unsigned char color;

#ifdef VECTORIZE
    #pragma omp simd
#endif
    for(col = 0; col < WIDTH; col++)
    {
	x0 = coords.l + (coords.r-coords.l)*((float_t)col/(float_t)WIDTH);
	y0 = coords.b + (coords.t-coords.b)*((float_t)row/(float_t)HEIGHT);

	iteration = mandel_iter(x0, y0);

	color = iteration % NCOLORS;
	img->R[HEIGHT-1-row][col] = palette[color][0];
	img->G[HEIGHT-1-row][col] = palette[color][1];
	img->B[HEIGHT-1-row][col] = palette[color][2];
    }
}


// Modules /////////////////////////////////////////////////////////////


SC_MODULE(Monitor)
{
    sc_fifo_in<IMAGE> ImgIn;
    IMAGE image;        // current received image

    int SaveImage(const char *Filename,
                unsigned char R[HEIGHT][WIDTH],
                unsigned char G[HEIGHT][WIDTH],
                unsigned char B[HEIGHT][WIDTH])
    {
#ifdef JPEG
    char    SysCmd[256];
#endif /* JPEG */
    FILE    *File;
    int     y, x;

    assert(Filename);
    assert(strlen(Filename) < 80);
#ifdef DEBUG
    printf("DEBUG: Saving file %s...\n", Filename);
#endif
    File = fopen(Filename, "w");
    if (!File)
    {
        printf("\nCannot open file \"%s\" for writing!\n", Filename);
        return 1;
    }
    fprintf(File, "P6\n");
    fprintf(File, "%d %d\n", WIDTH, HEIGHT);
    fprintf(File, "255\n");
    for(y=0; y<HEIGHT; y++)
    {
        for(x=0; x<WIDTH; x++)
        {
            fputc(R[y][x], File);
            fputc(G[y][x], File);
            fputc(B[y][x], File);
        }
    }
    if (ferror(File))
    {
        printf("\nFile error while writing to file!\n");
        fclose(File);
        return 2;
    }
    fclose(File);
#ifdef DEBUG
    printf("DEBUG: File %s was saved successfully.\n", Filename);
#endif
#ifdef JPEG
#ifdef DEBUG
    printf("DEBUG: Converting file %s to %s.jpg...\n", Filename, Filename);
#endif
    sprintf(SysCmd, "pnmtojpeg %s >%s.jpg", Filename, Filename);
    if (system(SysCmd) != 0)
    {
        printf("\nError while converting PPM to JPG:\nCommand \"%s\" failed!\n", SysCmd);
        return 10;
    }
#endif /* JPEG */
    return 0;
    } /* end of SaveImage */

    void main(void)
    {
#if (MAX_IMAGE_FILES > 0)
        char outfilename[128];
        int n;
#endif
        int i;

        for(i=0; i<NUM_IMAGES; i++)
        {
            ImgIn.read(image);
#ifdef DEBUG
            printf("Monitor received image %d.\n", i);
#endif
#if (MAX_IMAGE_FILES > 0)
            n = i % MAX_IMAGE_FILES;
            sprintf(outfilename, "mandelbrot_%03d.ppm", n);
            if (SaveImage(outfilename, image.R, image.G, image.B))
            {
                fprintf(stderr, "Error writing the image \"%s\".\n", outfilename);
                exit(10);
            }
#endif
#ifdef DISPLAY
            if (_SDL_SysC_DisplayClosed())
            {
#ifdef DEBUG
                printf("Display closed, monitor quits.\n");
#endif
                sc_stop();
            }
#endif /* DISPLAY */
        }
#ifdef DEBUG
        printf("Monitor quits.\n");
#endif
        sc_stop();
    }

    SC_CTOR(Monitor)
    {
        SC_THREAD(main);
        set_stack_size(128*1024*1024);
    }
};


SC_MODULE(Stimulus)
{
    sc_fifo_out<COORDS> CoordsOut;
    COORDS coords;

    void main(void)
    {
#ifdef DISPLAY
        char title[] = "Mandelbrot XXXX";
#endif /* DISPLAY */
        int i = 0;
        float_t center_x, center_y;
#ifdef LINEAR_MOTION
        float_t offset_x, offset_y;
#endif
#ifdef FASTER_MOTION
        float_t start_x, start_y;
        float_t dist_x, dist_y;
        float_t reverse_x, reverse_y;
        float_t factor_x, factor_y;
        float_t product_x, product_y;
#endif
#ifdef SLOWER_MOTION
        float_t stop_x, stop_y;
        float_t dist_x, dist_y;
        float_t reverse_x, reverse_y;
        float_t divisor_x, divisor_y;
        float_t quotient_x, quotient_y;
#endif
#ifdef DEBUG
        float_t prev_x, prev_y;
#endif
        float_t width, height;

        coords.l = START_LEFT;
        coords.r = START_RIGHT;
        coords.b = START_BOTTOM;
        coords.t = START_TOP;
        center_x = coords.l + (coords.r - coords.l)/2.0;
        center_y = coords.b + (coords.t - coords.b)/2.0;
#ifdef DEBUG
        printf("DEBUG: 1st coordinates (%+1.9f,%+1.9f)\n",
                (double)center_x, (double)center_y);
        prev_x = center_x;
        prev_y = center_y;
#endif
#ifdef LINEAR_MOTION
        offset_x = (TARGET_COORDS_X - center_x)/((float_t)NUM_IMAGES);
        offset_y = (TARGET_COORDS_Y - center_y)/((float_t)NUM_IMAGES);
#endif
#ifdef FASTER_MOTION
        start_x = center_x;
        dist_x = TARGET_COORDS_X - start_x;
        if (dist_x >= 0)
            reverse_x = 1.0;
        else
            reverse_x = -1.0;
        factor_x = exp(log(dist_x * reverse_x + 1.0)/((float_t)NUM_IMAGES));
        product_x = 1.0;
        start_y = center_y;
        dist_y = TARGET_COORDS_Y - start_y;
        if (dist_y >= 0)
            reverse_y = 1.0;
        else
            reverse_y = -1.0;
        factor_y = exp(log(dist_y * reverse_y + 1.0)/((float_t)NUM_IMAGES));
        product_y = 1.0;
#endif
#ifdef SLOWER_MOTION
        stop_x = TARGET_COORDS_X;
        dist_x = TARGET_COORDS_X - center_x;
        if (dist_x >= 0)
            reverse_x = 1.0;
        else
            reverse_x = -1.0;
        divisor_x = exp(log(dist_x * reverse_x + 1.0)/((float_t)NUM_IMAGES));
        quotient_x = dist_x * reverse_x + 1.0;
        stop_y = TARGET_COORDS_Y;
        dist_y = TARGET_COORDS_Y - center_y;
        if (dist_y >= 0)
            reverse_y = 1.0;
        else
            reverse_y = -1.0;
        divisor_y = exp(log(dist_y * reverse_y + 1.0)/((float_t)NUM_IMAGES));
        quotient_y = dist_y * reverse_y + 1.0;
#endif

#ifdef DISPLAY
        sprintf(title, "Mandelbrot %d", PAR);
        _SDL_SysC_InitDisplay(WIDTH, HEIGHT, title);
#endif /* DISPLAY */

        for(i=0; i<NUM_IMAGES; i++)
        {
#ifdef DEBUG
            printf("Stimulus sends coordinates set %d.\n", i);
#endif
            CoordsOut.write(coords);

#ifdef LINEAR_MOTION
            center_x += offset_x;
            center_y += offset_y;
#endif
#ifdef FASTER_MOTION
            product_x *= factor_x;
            product_y *= factor_y;
            center_x = start_x + (product_x-1.0) * reverse_x;
            center_y = start_y + (product_y-1.0) * reverse_y;
#endif
#ifdef SLOWER_MOTION
            quotient_x /= divisor_x;
            quotient_y /= divisor_y;
            center_x = stop_x - (quotient_x-1.0) * reverse_x;
            center_y = stop_y - (quotient_y-1.0) * reverse_y;
#endif
#ifdef DEBUG
            printf("DEBUG: Add coordinates (%+1.9f,%+1.9f)\n",
                    (double)(center_x-prev_x), (double)(center_y-prev_y));
            printf("DEBUG: New coordinates (%+1.9f,%+1.9f)\n",
                    (double)center_x, (double)center_y);
            prev_x = center_x;
            prev_y = center_y;
#endif
            width = (coords.r - coords.l) * (TARGET_ZOOM/2.0);
            height = (coords.t - coords.b) * (TARGET_ZOOM/2.0);
            coords.l = center_x - width;
            coords.r = center_x + width;
            coords.b = center_y - height;
            coords.t = center_y + height;
        }
    }

    SC_CTOR(Stimulus)
    {
        SC_THREAD(main);
    }
};


SC_MODULE(DUT)
{
    sc_fifo_in<COORDS> CoordsIn;
    sc_fifo_out<IMAGE> ImgOut;

    COORDS coords;
    IMAGE image;

    sc_event mandelbrot_run;
    sc_event_and_list main_continue;

#define M(n) sc_event ready_mandel##n
    M(1);
#if PAR > 1
    M(2);
#endif
#if PAR > 2
    M(3); M(4);
#endif
#if PAR > 4
    M(5); M(6); M(7); M(8);
#endif
#if PAR > 8
    M(9); M(10); M(11); M(12); M(13); M(14); M(15); M(16);
#endif
#if PAR > 16
    M(17); M(18); M(19); M(20); M(21); M(22); M(23); M(24);
    M(25); M(26); M(27); M(28); M(29); M(30); M(31); M(32);
#endif
#if PAR > 32
    M(33); M(34); M(35); M(36); M(37); M(38); M(39); M(40);
    M(41); M(42); M(43); M(44); M(45); M(46); M(47); M(48);
    M(49); M(50); M(51); M(52); M(53); M(54); M(55); M(56);
    M(57); M(58); M(59); M(60); M(61); M(62); M(63); M(64);
#endif
#if PAR > 64
    M(65); M(66); M(67); M(68); M(69); M(70); M(71); M(72);
    M(73); M(74); M(75); M(76); M(77); M(78); M(79); M(80);
    M(81); M(82); M(83); M(84); M(85); M(86); M(87); M(88);
    M(89); M(90); M(91); M(92); M(93); M(94); M(95); M(96);
    M(97); M(98); M(99); M(100); M(101); M(102); M(103); M(104);
    M(105); M(106); M(107); M(108); M(109); M(110); M(111); M(112);
    M(113); M(114); M(115); M(116); M(117); M(118); M(119); M(120);
    M(121); M(122); M(123); M(124); M(125); M(126); M(127); M(128);
#endif
#if PAR > 128
    M(129); M(130); M(131); M(132); M(133); M(134); M(135); M(136);
    M(137); M(138); M(139); M(140); M(141); M(142); M(143); M(144);
    M(145); M(146); M(147); M(148); M(149); M(150); M(151); M(152);
    M(153); M(154); M(155); M(156); M(157); M(158); M(159); M(160);
    M(161); M(162); M(163); M(164); M(165); M(166); M(167); M(168);
    M(169); M(170); M(171); M(172); M(173); M(174); M(175); M(176);
    M(177); M(178); M(179); M(180); M(181); M(182); M(183); M(184);
    M(185); M(186); M(187); M(188); M(189); M(190); M(191); M(192);
    M(193); M(194); M(195); M(196); M(197); M(198); M(199); M(200);
    M(201); M(202); M(203); M(204); M(205); M(206); M(207); M(208);
    M(209); M(210); M(211); M(212); M(213); M(214); M(215); M(216);
    M(217); M(218); M(219); M(220); M(221); M(222); M(223); M(224);
    M(225); M(226); M(227); M(228); M(229); M(230); M(231); M(232);
    M(233); M(234); M(235); M(236); M(237); M(238); M(239); M(240);
    M(241); M(242); M(243); M(244); M(245); M(246); M(247); M(248);
    M(249); M(250); M(251); M(252); M(253); M(254); M(255); M(256);
#endif
#undef M

// The following four versions of mandlebrot##n are almost the same,
// differing in combination of the following two conditions:
// - USE_SC_THREAD: an infinite top-level loop and wait() vs. next_trigger()
// - DISPLAY: invoke SDL APIs vs. not
// Note that DISPLAY must be off for SIMD vectorization (RD, 05/25/16)

#ifdef USE_SC_THREAD

#ifdef DISPLAY
#define MANDELBROT(n, Coords, rowStart, rowEnd, Image)                                          \
void mandelbrot##n (void)                                                                       \
{                                                                                               \
    float_t x, y;                                                                               \
    float_t x0, y0;                                                                             \
    float_t xtemp;                                                                              \
    unsigned int iteration;                                                                     \
    const unsigned int max_iteration = MAX_ITER;                                                \
    unsigned char color;                                                                        \
    int row, col;                                                                               \
                                                                                                \
    while(true)                                                                                 \
    {                                                                                           \
        /* The following is taken (with very few adaptations) from: */                          \
        /* http://en.wikipedia.org/wiki/Mandelbrot_set#For_programmers*/                        \
        /* RD, 10/27/12. */                                                                     \
                                                                                                \
        wait(mandelbrot_run);                                                                   \
        /* For each pixel on the screen do: */                                                  \
        for(row = rowStart; row < rowEnd; row++)                                                \
        {                                                                                       \
            for(col = 0; col < WIDTH; col++)                                                    \
            {                                                                                   \
                /* x0 = scaled x coordinate of pixel */                                         \
                /*      (must be scaled to lie somewhere in the mandelbrot X scale (-2.5, 1) */ \
                x0 = Coords.l + (Coords.r-Coords.l)*((float_t)col/(float_t)WIDTH);              \
                /* y0 = scaled y coordinate of pixel */                                         \
                /*      (must be scaled to lie somewhere in the mandelbrot Y scale (-1, 1) */   \
                y0 = Coords.b + (Coords.t-Coords.b)*((float_t)row/(float_t)HEIGHT);             \
                                                                                                \
                x = 0;                                                                          \
                y = 0;                                                                          \
                                                                                                \
                iteration = 0;                                                                  \
                /* max_iteration = 1000; */                                                     \
                                                                                                \
                while ( x*x + y*y < 2*2  &&  iteration < max_iteration )                        \
                {                                                                               \
                    xtemp = x*x - y*y + x0;                                                     \
                    y = 2*x*y + y0;                                                             \
                                                                                                \
                    x = xtemp;                                                                  \
                                                                                                \
                    iteration = iteration + 1;                                                  \
                }                                                                               \
                                                                                                \
                /* color = iteration */                                                         \
                color = iteration % NCOLORS;                                                    \
                                                                                                \
                /* plot(x0,y0,color) */                                                         \
                Image.R[HEIGHT-1-row][col] = palette[color][0];                                 \
                Image.G[HEIGHT-1-row][col] = palette[color][1];                                 \
                Image.B[HEIGHT-1-row][col] = palette[color][2];                                 \
                                                                                                \
                _SDL_SysC_DrawPixel(palette[color][0], palette[color][1],                       \
                                    palette[color][2], col, HEIGHT-1-row);                      \
            }                                                                                   \
            _SDL_SysC_UpdateRect(0, HEIGHT-1-row, WIDTH, 1);                                    \
        }                                                                                       \
                                                                                                \
        ready_mandel##n.notify(SC_ZERO_TIME);                                                   \
    }                                                                                           \
}
#else
#define MANDELBROT(n, Coords, rowStart, rowEnd, Image)                                          \
void mandelbrot##n (void)                                                                       \
{                                                                                               \
    int row;                                                                                    \
                                                                                                \
    while(true)                                                                                 \
    {                                                                                           \
        wait(mandelbrot_run);                                                                   \
                                                                                                \
        for(row = rowStart; row < rowEnd; row++)                                                \
        {                                                                                       \
            mandel_row(&image, coords, row); /* SIMD optimized Mandelbrot for one row */        \
        }                                                                                       \
                                                                                                \
        ready_mandel##n.notify(SC_ZERO_TIME);                                                   \
    }                                                                                           \
}
#endif /* DISPLAY */

#else /* Use SC_METHOD */

#ifdef DISPLAY
#define MANDELBROT(n, Coords, rowStart, rowEnd, Image)                                          \
void mandelbrot##n (void)                                                                       \
{                                                                                               \
    float_t x, y;                                                                               \
    float_t x0, y0;                                                                             \
    float_t xtemp;                                                                              \
    unsigned int iteration;                                                                     \
    const unsigned int max_iteration = MAX_ITER;                                                \
    unsigned char color;                                                                        \
    int row, col;                                                                               \
                                                                                                \
    /* The following is taken (with very few adaptations) from: */                              \
    /* http://en.wikipedia.org/wiki/Mandelbrot_set#For_programmers*/                            \
    /* RD, 10/27/12. */                                                                         \
                                                                                                \
    /* For each pixel on the screen do: */                                                      \
    for(row = rowStart; row < rowEnd; row++)                                                    \
    {                                                                                           \
        for(col = 0; col < WIDTH; col++)                                                        \
        {                                                                                       \
            /* x0 = scaled x coordinate of pixel */                                             \
            /*      (must be scaled to lie somewhere in the mandelbrot X scale (-2.5, 1) */     \
            x0 = Coords.l + (Coords.r-Coords.l)*((float_t)col/(float_t)WIDTH);                  \
            /* y0 = scaled y coordinate of pixel */                                             \
            /*      (must be scaled to lie somewhere in the mandelbrot Y scale (-1, 1) */       \
            y0 = Coords.b + (Coords.t-Coords.b)*((float_t)row/(float_t)HEIGHT);                 \
                                                                                                \
            x = 0;                                                                              \
            y = 0;                                                                              \
                                                                                                \
            iteration = 0;                                                                      \
            /* max_iteration = 1000; */                                                         \
                                                                                                \
            while ( x*x + y*y < 2*2  &&  iteration < max_iteration )                            \
            {                                                                                   \
                xtemp = x*x - y*y + x0;                                                         \
                y = 2*x*y + y0;                                                                 \
                                                                                                \
                x = xtemp;                                                                      \
                                                                                                \
                iteration = iteration + 1;                                                      \
            }                                                                                   \
                                                                                                \
            /* color = iteration */                                                             \
            color = iteration % NCOLORS;                                                        \
                                                                                                \
            /* plot(x0,y0,color) */                                                             \
            Image.R[HEIGHT-1-row][col] = palette[color][0];                                     \
            Image.G[HEIGHT-1-row][col] = palette[color][1];                                     \
            Image.B[HEIGHT-1-row][col] = palette[color][2];                                     \
                                                                                                \
            _SDL_SysC_DrawPixel(palette[color][0], palette[color][1],                           \
                                palette[color][2], col, HEIGHT-1-row);                          \
        }                                                                                       \
        _SDL_SysC_UpdateRect(0, HEIGHT-1-row, WIDTH, 1);                                        \
    }                                                                                           \
                                                                                                \
    next_trigger();                                                                             \
    ready_mandel##n.notify(SC_ZERO_TIME);                                                       \
}
#else
#define MANDELBROT(n, Coords, rowStart, rowEnd, Image)                                          \
void mandelbrot##n (void)                                                                       \
{                                                                                               \
    int row;                                                                                    \
                                                                                                \
    for(row = rowStart; row < rowEnd; row++)                                                    \
    {                                                                                           \
        mandel_row(row); /* SIMD optimized Mandelbrot for one row */                            \
    }                                                                                           \
                                                                                                \
    next_trigger();                                                                             \
    ready_mandel##n.notify(SC_ZERO_TIME);                                                       \
}
#endif /* DISPLAY */

#endif /* USE_SC_THREAD */

#define M(n) MANDELBROT(n, coords, (HEIGHT/PAR*(n-1)), (HEIGHT/PAR*n), image)
    M(1) 
#if PAR > 1
    M(2) 
#endif
#if PAR > 2
    M(3) M(4) 
#endif
#if PAR > 4
    M(5) M(6) M(7) M(8) 
#endif
#if PAR > 8
    M(9) M(10) M(11) M(12) M(13) M(14) M(15) M(16)
#endif
#if PAR > 16
    M(17) M(18) M(19) M(20) M(21) M(22) M(23) M(24)
    M(25) M(26) M(27) M(28) M(29) M(30) M(31) M(32)
#endif
#if PAR > 32
    M(33) M(34) M(35) M(36) M(37) M(38) M(39) M(40)
    M(41) M(42) M(43) M(44) M(45) M(46) M(47) M(48)
    M(49) M(50) M(51) M(52) M(53) M(54) M(55) M(56)
    M(57) M(58) M(59) M(60) M(61) M(62) M(63) M(64)
#endif
#if PAR > 64
    M(65) M(66) M(67) M(68) M(69) M(70) M(71) M(72)
    M(73) M(74) M(75) M(76) M(77) M(78) M(79) M(80)
    M(81) M(82) M(83) M(84) M(85) M(86) M(87) M(88)
    M(89) M(90) M(91) M(92) M(93) M(94) M(95) M(96)
    M(97) M(98) M(99) M(100) M(101) M(102) M(103) M(104)
    M(105) M(106) M(107) M(108) M(109) M(110) M(111) M(112)
    M(113) M(114) M(115) M(116) M(117) M(118) M(119) M(120)
    M(121) M(122) M(123) M(124) M(125) M(126) M(127) M(128)
#endif
#if PAR > 128
    M(129) M(130) M(131) M(132) M(133) M(134) M(135) M(136)
    M(137) M(138) M(139) M(140) M(141) M(142) M(143) M(144)
    M(145) M(146) M(147) M(148) M(149) M(150) M(151) M(152)
    M(153) M(154) M(155) M(156) M(157) M(158) M(159) M(160)
    M(161) M(162) M(163) M(164) M(165) M(166) M(167) M(168)
    M(169) M(170) M(171) M(172) M(173) M(174) M(175) M(176)
    M(177) M(178) M(179) M(180) M(181) M(182) M(183) M(184)
    M(185) M(186) M(187) M(188) M(189) M(190) M(191) M(192)
    M(193) M(194) M(195) M(196) M(197) M(198) M(199) M(200)
    M(201) M(202) M(203) M(204) M(205) M(206) M(207) M(208)
    M(209) M(210) M(211) M(212) M(213) M(214) M(215) M(216)
    M(217) M(218) M(219) M(220) M(221) M(222) M(223) M(224)
    M(225) M(226) M(227) M(228) M(229) M(230) M(231) M(232)
    M(233) M(234) M(235) M(236) M(237) M(238) M(239) M(240)
    M(241) M(242) M(243) M(244) M(245) M(246) M(247) M(248)
    M(249) M(250) M(251) M(252) M(253) M(254) M(255) M(256)
#endif
#undef M

#undef MANDELBROT

    void main(void)
    {
        while(1)
        {
            CoordsIn.read(coords);
#ifdef DISPLAY
            // _SDL_SysC_ClearDisplay();
            _SDL_SysC_FillDisplay(42,42,42);
#endif /* DISPLAY */
            mandelbrot_run.notify(SC_ZERO_TIME);
            wait(main_continue);
            ImgOut.write(image);
        }
    }

    void before_end_of_elaboration()
    {
        // initialize sc_event_and_list main_continue
        #define M(n) ready_mandel##n
                main_continue &= M(1);
        #if PAR >1
                main_continue &= M(2);
        #endif
        #if PAR > 2
                main_continue &= M(3);
                main_continue &= M(4);
        #endif
        #if PAR > 4
                main_continue &= M(5);
                main_continue &= M(6);
                main_continue &= M(7);
                main_continue &= M(8);
        #endif
        #if PAR > 8
                main_continue &= M(9);
                main_continue &= M(10);
                main_continue &= M(11);
                main_continue &= M(12);
                main_continue &= M(13);
                main_continue &= M(14);
                main_continue &= M(15);
                main_continue &= M(16);
        #endif
        #if PAR > 16
                main_continue &= M(17);
                main_continue &= M(18);
                main_continue &= M(19);
                main_continue &= M(20);
                main_continue &= M(21);
                main_continue &= M(22);
                main_continue &= M(23);
                main_continue &= M(24);
                main_continue &= M(25);
                main_continue &= M(26);
                main_continue &= M(27);
                main_continue &= M(28);
                main_continue &= M(29);
                main_continue &= M(30);
                main_continue &= M(31);
                main_continue &= M(32);
        #endif
        #if PAR > 32
                main_continue &= M(33);
                main_continue &= M(34);
                main_continue &= M(35);
                main_continue &= M(36);
                main_continue &= M(37);
                main_continue &= M(38);
                main_continue &= M(39);
                main_continue &= M(40);
                main_continue &= M(41);
                main_continue &= M(42);
                main_continue &= M(43);
                main_continue &= M(44);
                main_continue &= M(45);
                main_continue &= M(46);
                main_continue &= M(47);
                main_continue &= M(48);
                main_continue &= M(49);
                main_continue &= M(50);
                main_continue &= M(51);
                main_continue &= M(52);
                main_continue &= M(53);
                main_continue &= M(54);
                main_continue &= M(55);
                main_continue &= M(56);
                main_continue &= M(57);
                main_continue &= M(58);
                main_continue &= M(59);
                main_continue &= M(60);
                main_continue &= M(61);
                main_continue &= M(62);
                main_continue &= M(63);
                main_continue &= M(64);
        #endif
        #if PAR > 64
                main_continue &= M(65);
                main_continue &= M(66);
                main_continue &= M(67);
                main_continue &= M(68);
                main_continue &= M(69);
                main_continue &= M(70);
                main_continue &= M(71);
                main_continue &= M(72);
                main_continue &= M(73);
                main_continue &= M(74);
                main_continue &= M(75);
                main_continue &= M(76);
                main_continue &= M(77);
                main_continue &= M(78);
                main_continue &= M(79);
                main_continue &= M(80);
                main_continue &= M(81);
                main_continue &= M(82);
                main_continue &= M(83);
                main_continue &= M(84);
                main_continue &= M(85);
                main_continue &= M(86);
                main_continue &= M(87);
                main_continue &= M(88);
                main_continue &= M(89);
                main_continue &= M(90);
                main_continue &= M(91);
                main_continue &= M(92);
                main_continue &= M(93);
                main_continue &= M(94);
                main_continue &= M(95);
                main_continue &= M(96);
                main_continue &= M(97);
                main_continue &= M(98);
                main_continue &= M(99);
                main_continue &= M(100);
                main_continue &= M(101);
                main_continue &= M(102);
                main_continue &= M(103);
                main_continue &= M(104);
                main_continue &= M(105);
                main_continue &= M(106);
                main_continue &= M(107);
                main_continue &= M(108);
                main_continue &= M(109);
                main_continue &= M(110);
                main_continue &= M(111);
                main_continue &= M(112);
                main_continue &= M(113);
                main_continue &= M(114);
                main_continue &= M(115);
                main_continue &= M(116);
                main_continue &= M(117);
                main_continue &= M(118);
                main_continue &= M(119);
                main_continue &= M(120);
                main_continue &= M(121);
                main_continue &= M(122);
                main_continue &= M(123);
                main_continue &= M(124);
                main_continue &= M(125);
                main_continue &= M(126);
                main_continue &= M(127);
                main_continue &= M(128);
        #endif
        #if PAR > 128
                main_continue &= M(129);
                main_continue &= M(130);
                main_continue &= M(131);
                main_continue &= M(132);
                main_continue &= M(133);
                main_continue &= M(134);
                main_continue &= M(135);
                main_continue &= M(136);
                main_continue &= M(137);
                main_continue &= M(138);
                main_continue &= M(139);
                main_continue &= M(140);
                main_continue &= M(141);
                main_continue &= M(142);
                main_continue &= M(143);
                main_continue &= M(144);
                main_continue &= M(145);
                main_continue &= M(146);
                main_continue &= M(147);
                main_continue &= M(148);
                main_continue &= M(149);
                main_continue &= M(150);
                main_continue &= M(151);
                main_continue &= M(152);
                main_continue &= M(153);
                main_continue &= M(154);
                main_continue &= M(155);
                main_continue &= M(156);
                main_continue &= M(157);
                main_continue &= M(158);
                main_continue &= M(159);
                main_continue &= M(160);
                main_continue &= M(161);
                main_continue &= M(162);
                main_continue &= M(163);
                main_continue &= M(164);
                main_continue &= M(165);
                main_continue &= M(166);
                main_continue &= M(167);
                main_continue &= M(168);
                main_continue &= M(169);
                main_continue &= M(170);
                main_continue &= M(171);
                main_continue &= M(172);
                main_continue &= M(173);
                main_continue &= M(174);
                main_continue &= M(175);
                main_continue &= M(176);
                main_continue &= M(177);
                main_continue &= M(178);
                main_continue &= M(179);
                main_continue &= M(180);
                main_continue &= M(181);
                main_continue &= M(182);
                main_continue &= M(183);
                main_continue &= M(184);
                main_continue &= M(185);
                main_continue &= M(186);
                main_continue &= M(187);
                main_continue &= M(188);
                main_continue &= M(189);
                main_continue &= M(190);
                main_continue &= M(191);
                main_continue &= M(192);
                main_continue &= M(193);
                main_continue &= M(194);
                main_continue &= M(195);
                main_continue &= M(196);
                main_continue &= M(197);
                main_continue &= M(198);
                main_continue &= M(199);
                main_continue &= M(200);
                main_continue &= M(201);
                main_continue &= M(202);
                main_continue &= M(203);
                main_continue &= M(204);
                main_continue &= M(205);
                main_continue &= M(206);
                main_continue &= M(207);
                main_continue &= M(208);
                main_continue &= M(209);
                main_continue &= M(210);
                main_continue &= M(211);
                main_continue &= M(212);
                main_continue &= M(213);
                main_continue &= M(214);
                main_continue &= M(215);
                main_continue &= M(216);
                main_continue &= M(217);
                main_continue &= M(218);
                main_continue &= M(219);
                main_continue &= M(220);
                main_continue &= M(221);
                main_continue &= M(222);
                main_continue &= M(223);
                main_continue &= M(224);
                main_continue &= M(225);
                main_continue &= M(226);
                main_continue &= M(227);
                main_continue &= M(228);
                main_continue &= M(229);
                main_continue &= M(230);
                main_continue &= M(231);
                main_continue &= M(232);
                main_continue &= M(233);
                main_continue &= M(234);
                main_continue &= M(235);
                main_continue &= M(236);
                main_continue &= M(237);
                main_continue &= M(238);
                main_continue &= M(239);
                main_continue &= M(240);
                main_continue &= M(241);
                main_continue &= M(242);
                main_continue &= M(243);
                main_continue &= M(244);
                main_continue &= M(245);
                main_continue &= M(246);
                main_continue &= M(247);
                main_continue &= M(248);
                main_continue &= M(249);
                main_continue &= M(250);
                main_continue &= M(251);
                main_continue &= M(252);
                main_continue &= M(253);
                main_continue &= M(254);
                main_continue &= M(255);
                main_continue &= M(256);
        #endif
        #undef M
        // end of initialization of main_continue
    }

    SC_CTOR(DUT)
    {
        // construct main thread
        SC_THREAD(main);

        // construct worker methods
#ifdef USE_SC_THREAD
#define M(n) SC_THREAD(mandelbrot##n); sensitive << mandelbrot_run;
#else
#define M(n) SC_METHOD(mandelbrot##n); sensitive << mandelbrot_run;
#endif /* USE_SC_THREAD */
        M(1);
#if PAR > 1
        M(2);
#endif
#if PAR > 2
        M(3); M(4);
#endif
#if PAR > 4
        M(5); M(6); M(7); M(8);
#endif
#if PAR > 8
        M(9); M(10); M(11); M(12); M(13); M(14); M(15); M(16);
#endif
#if PAR > 16
        M(17); M(18); M(19); M(20); M(21); M(22); M(23); M(24);
        M(25); M(26); M(27); M(28); M(29); M(30); M(31); M(32);
#endif
#if PAR > 32
        M(33); M(34); M(35); M(36); M(37); M(38); M(39); M(40);
        M(41); M(42); M(43); M(44); M(45); M(46); M(47); M(48);
        M(49); M(50); M(51); M(52); M(53); M(54); M(55); M(56);
        M(57); M(58); M(59); M(60); M(61); M(62); M(63); M(64);
#endif
#if PAR > 64
        M(65); M(66); M(67); M(68); M(69); M(70); M(71); M(72);
        M(73); M(74); M(75); M(76); M(77); M(78); M(79); M(80);
        M(81); M(82); M(83); M(84); M(85); M(86); M(87); M(88);
        M(89); M(90); M(91); M(92); M(93); M(94); M(95); M(96);
        M(97); M(98); M(99); M(100); M(101); M(102); M(103); M(104);
        M(105); M(106); M(107); M(108); M(109); M(110); M(111); M(112);
        M(113); M(114); M(115); M(116); M(117); M(118); M(119); M(120);
        M(121); M(122); M(123); M(124); M(125); M(126); M(127); M(128);
#endif
#if PAR > 128
        M(129); M(130); M(131); M(132); M(133); M(134); M(135); M(136);
        M(137); M(138); M(139); M(140); M(141); M(142); M(143); M(144);
        M(145); M(146); M(147); M(148); M(149); M(150); M(151); M(152);
        M(153); M(154); M(155); M(156); M(157); M(158); M(159); M(160);
        M(161); M(162); M(163); M(164); M(165); M(166); M(167); M(168);
        M(169); M(170); M(171); M(172); M(173); M(174); M(175); M(176);
        M(177); M(178); M(179); M(180); M(181); M(182); M(183); M(184);
        M(185); M(186); M(187); M(188); M(189); M(190); M(191); M(192);
        M(193); M(194); M(195); M(196); M(197); M(198); M(199); M(200);
        M(201); M(202); M(203); M(204); M(205); M(206); M(207); M(208);
        M(209); M(210); M(211); M(212); M(213); M(214); M(215); M(216);
        M(217); M(218); M(219); M(220); M(221); M(222); M(223); M(224);
        M(225); M(226); M(227); M(228); M(229); M(230); M(231); M(232);
        M(233); M(234); M(235); M(236); M(237); M(238); M(239); M(240);
        M(241); M(242); M(243); M(244); M(245); M(246); M(247); M(248);
        M(249); M(250); M(251); M(252); M(253); M(254); M(255); M(256);
#endif
#undef M
    }
};


SC_MODULE(DataIn)
{
    sc_fifo_in<COORDS> CoordsIn;
    sc_fifo_out<COORDS> CoordsOut;

    COORDS coords;

    void main(void)
    {
        while(1)
        {
            CoordsIn.read(coords);
            CoordsOut.write(coords);
        }
    }

    SC_CTOR(DataIn)
    {
        SC_THREAD(main);
    }
};


SC_MODULE(DataOut)
{
    sc_fifo_in<IMAGE> ImgIn;
    sc_fifo_out<IMAGE> ImgOut;

    IMAGE image;

    void main(void)
    {
        while(1)
        {
            ImgIn.read(image);
            ImgOut.write(image);
        }
    }

    SC_CTOR(DataOut)
    {
        SC_THREAD(main);
        set_stack_size(128*1024*1024);
    }
};


SC_MODULE(Platform)
{
    sc_fifo_in<COORDS> CoordsIn;
    sc_fifo_out<IMAGE> ImgOut;

    sc_fifo<COORDS>   platform_c1;
    sc_fifo<IMAGE>    platform_c2;
    DataIn          din;
    DUT             dut;
    DataOut         dout;

    SC_CTOR(Platform):
        platform_c1("platform_c1"),
        platform_c2("platform_c2"),
        din("din"),
        dut("dut"),
        dout("dout")
    {
    }

    void before_end_of_elaboration()
    {
        din.CoordsIn.bind(CoordsIn);
        din.CoordsOut.bind(platform_c1);

        dut.CoordsIn.bind(platform_c1);
        dut.ImgOut.bind(platform_c2);

        dout.ImgIn.bind(platform_c2);
        dout.ImgOut.bind(ImgOut);
    }
};


SC_MODULE(TOP)
{
    sc_fifo<COORDS>   c1;
    sc_fifo<IMAGE>    c2;

    Stimulus    stimulus;
    Platform    platform;
    Monitor     monitor;

    SC_CTOR(TOP):
        c1("c1"),
        c2("c2"),
        stimulus("stimulus"),
        platform("platform"),
        monitor("monitor")
    {
    }

    void before_end_of_elaboration()
    {
        stimulus.CoordsOut.bind(c1);

        platform.CoordsIn.bind(c1);
        platform.ImgOut.bind(c2);

        monitor.ImgIn.bind(c2);
    }
};


int sc_main(int argc, char* argv[])
{
    TOP top("top");
#ifdef DEBUG
    printf("Mandelbrot starts.\n");
#endif
    sc_start();
#ifdef DEBUG
    printf("Mandelbrot done.\n");
#endif
    return 0;
}

// EOF
