#include <iostream>
#include <string>

#include "risc.h"

#include <libgen.h>

risc::tools::ParseStatus status;

std::string parse_out_elab_filename();
std::string get_design_filename(int argc, char **argv);

#define HELP_PAGE                                                                   \
"risc: "RISC_LIBRARY_NAME" Dynamic Elaborator, V"RISC_VERSION_NUMBER"\n"            \
"      "RISC_COPYRIGHT"\n"                                                          \
"\n"                                                                                \
"Usage: elab [<option>...] <design> [<option>...]\n"                                \
"\n"                                                                                \
"Options:               Description:                    Default:\n"                 \
"-h|--help              print this usage info, quit     none\n"                     \
"-v|--verbose           increment verbosity             silent\n"                   \
"-vv                    increment verbosity by 2        silent\n"                   \
"-vvv                   increment verbosity by 3        silent\n"                   \
"-w|--warnings          increment warning level         no warnings\n"              \
"-ww                    increment warning level by 2    no warnings\n"              \
"-www                   increment warning level by 3    no warnings\n"              \
"-g                     produce debugging information   no debugging infos\n"       \
"-O|-O<level>           enable optimizations            no optimization\n"          \
"-I<dir>                add dir to include path         $SYSTEMC_LW_HOME/include\n" \
"-L<dir>                add dir to library path         $SYSTEMC_OOP_HOME/lib\n"    \
"-l<lib>                add lib to linker libraries     -lsystemc\n"                \
"-c                     compile only, do not link       compile and link\n"         \
"-o <filename>          name output file as filename    a.out\n"                    \
"-elab:o <filename>     export elaboration result       <design>.elab\n"            \
"<rose:option>          passed to ROSE compiler         none\n"                     \
"<g++_option>           passed to GNU C++ compiler      none\n"                     \
"\n"

std::string parse_out_elab_filename(int &argc, char **&argv)
{
  for(int i=0; i < argc-1; i++) {

    if(strcmp("-elab:o", argv[i]) == 0) {
      std::string filename(argv[i+1]);

      // Here we are deleting the arguments '-inst:output_filename'
      std::vector<std::string> arguments =
        CommandlineProcessing::generateArgListFromArgcArgv(argc, argv);

      char **temp = NULL;

      arguments.erase(arguments.begin() + i, arguments.begin() + (i + 2));
      CommandlineProcessing::generateArgcArgvFromList(arguments,argc,temp);
      argv = temp;

      if(status.verbosityLevel > 0) {
        std::cout << "-elab:o filename: " << filename << std::endl;
      }

      return filename;
    }
  }

  std::string temp(get_design_filename(argc, argv));
  boost::filesystem::path p(temp);
  std::string design_name = p.filename().string();
  risc::tools::find_and_replace_all(design_name, "cpp", "elab");

  if(status.verbosityLevel > 0) {
    std::cout << "-elab:o filename (default): "
              << p.filename().string()
              << std::endl;
  }

  return design_name;
}

std::string get_design_filename(int argc, char **argv)
{
  if(argc > 1) {
    return argv[1];
  } else {
    std::cout << "No input design provided" << std::endl;
    exit(0);
  }
}


bool has_help_option(int argc, char* argv[])
{
  for(int i=0; i < argc; i++) {

    if(strcmp("-h", argv[i]) == 0 || strcmp("--help", argv[i]) == 0) {

      return true;
    }
  }

  // the option -h/--help was not set
  return false;
}


int main(int argc, char* argv[])
{
  std::string lw_include_path;
  std::string lw_library_path;
  lw_include_path = "-I";
  lw_include_path += getenv(SYSTEMC_LW_HOME_ENV);
  lw_include_path += "/include";
  lw_library_path = "-L";
  lw_library_path += getenv(SYSTEMC_LW_HOME_ENV);
  lw_library_path += "/lib-linux64";


  // add -I$SYSTEMC_LW_HOME/include
  risc::tools::AddCommandlineOption(argc, argv, lw_include_path);

  // add -L$SYSTEMC_LW_HOME/lib-linux64
  risc::tools::AddCommandlineOption(argc, argv, lw_library_path);

  // add -lsystemc
  risc::tools::AddCommandlineOption(argc, argv, "-lsystemc");

  risc::tools::ParseCommandlineOptions(argc, argv, &status);

  if(has_help_option(argc, argv)) {
    //print_help();
    std::cout << HELP_PAGE;
    return 0;
  }

  std::string filename = parse_out_elab_filename(argc, argv);

  if(status.verbosityLevel > 0) {
    std::cout << "start processing file ..." << std::endl;
  }

  if(status.verbosityLevel > 1) {
    std::cout << "output filename: " << filename << std::endl;
  }

  risc::Design d(argc,argv);
  risc::inst::instrument_reference_printing(filename);

  if(status.verbosityLevel > 0) {
    std::cout << "end processing file ..." << std::endl;
  }

  risc::tools::AddCommandlineOption(d, "-Xlinker");
  risc::tools::AddCommandlineOption(d, "-R");
  risc::tools::AddCommandlineOption(d, "-Xlinker");
  risc::tools::AddCommandlineOption(d, lw_library_path.erase(0, 2));

  risc::inst::risc_unparse(d, &status, NULL, NULL);

  risc::inst::risc_patch(d, &status,
      "sed -r -i \'s/create_method_process\\s*\\((.*)\\,(.*)\\,"
      "(.*)\\,(.*)\\,(.*)\\)/create_method_process "
      "(\\1, \\2, static_cast<sc_core::SC_ENTRY_FUNC>(\\3), \\4, "
      "\\5)/\' "RISC_UNPARSE_FILE_PREFIX"%s",
      "sed -r -i \'s/create_thread_process\\s*\\((.*)\\,(.*)\\,"
      "(.*)\\,(.*)\\,(.*)\\)/create_thread_process "
      "(\\1, \\2, static_cast<sc_core::SC_ENTRY_FUNC>(\\3), \\4, "
      "\\5)/\' "RISC_UNPARSE_FILE_PREFIX"%s",
      "sed -r -i \'s/create_cthread_process\\s*\\((.*)\\,(.*)\\,"
      "(.*)\\,(.*)\\,(.*)\\)/create_cthread_process "
      "(\\1, \\2, static_cast<sc_core::SC_ENTRY_FUNC>(\\3), \\4, "
      "\\5)/\' "RISC_UNPARSE_FILE_PREFIX"%s");

  risc::inst::risc_compile(d, &status);

  return 0;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
