#ifndef MODULE_H_INCLUDED_
#define MODULE_H_INCLUDED_

#include<iostream>
#include<string>
#include<vector>

#include"rose.h"
#include "class.h"

namespace risc {

class CThread;
class Function;
class HierarchicalChannelInstance;
class Method;
class Module;
class ModuleInstance;
class Port;
class PrimitiveChannel;
class PrimitiveChannelInstance;
class Thread;
class VariableInstance;

typedef std::pair<Port*, SgBoolValExp*> ResetSignal;

typedef std::vector<CThread*> CThreadVector;
typedef std::vector<Function*> FunctionVector;
typedef std::vector<HierarchicalChannelInstance*> HierarchicalChannelInstanceVector;
typedef std::vector<Method*> MethodVector;
typedef std::vector<Module*> ModuleVector;
typedef std::vector<Module*>::const_iterator ModuleVectorConstIter;
typedef std::vector<Module*>::iterator ModuleVectorIter;
typedef std::vector<ModuleInstance*> ModuleInstanceVector;
typedef std::vector<PrimitiveChannel*> PrimitiveChannelVector;
typedef std::vector<PrimitiveChannelInstance*> PrimitiveChannelInstanceVector;
typedef std::vector<ResetSignal*> ResetSignalVector;
typedef std::vector<Thread*> ThreadVector;
typedef std::vector<VariableInstance*> VariableInstanceVector;

/**
 *  \class Module
 *  \brief This class represents a definition of a module
 *
 *  A module is defined by the assembling of modules, channels, ports,
 *  processes, and others.
 */
class Module: public Class {

public:

  /**
   * Default constructor
   *
   * \brief Creates an empty module definition. All the lists of instances
   *        and sub definitions are empty.
   */
  explicit Module(SgClassDefinition *ast_node);

  /**
   * Default constructor
   *
   * \brief This constructor will be used for derived classes from sc_module
   *        Currently, this is only sc_channel
   */
  explicit Module(SgClassDefinition *ast_node, Type type);

  ~Module();

  /*
   * initialization functions
   */

  /**
   * \brief  This function searches for all functions, SC_THREADS, SC_CTHREADS,
   *         and SC_METHODS  in the module definition. The corresponding
   *         variable vectors will be used.
   */
  virtual void initialize_functions();

  /*
   * helper functions
   */

  /**
   * \brief  Prints the complete module design into the terminal
   */
  void print_module_definition();

  /**
   * \brief  Prints the hierarchy. Roots is this definition
   */
  void print_hierarchy(unsigned indent_level = 0);

  /*
   * Vectors module definition elements
   */

  /**
   * \brief Vector of threads which are defined in this module
   */
  ThreadVector threads_;

  /**
   * \brief Vector of cthreads which are defined in this module
   */
  CThreadVector cthreads_;

  /**
   * \brief Vector of methods which are defined in this module
   */
  MethodVector methods_;

  /**
   * \brief Vector of reset signals which are registered in this module
   */
  ResetSignalVector reset_signals_;

  /**
   * \brief Vector of async reset signals which are registered in this module
   */
  ResetSignalVector async_reset_signals_;

  /**
   * \brief This variable counts how many instances of this given type exists
   *        in the design
   */
  int instance_counter_;

  /**
   * \fn       get_mangled_type_name
   * \brief    This function returns the manged type name of the module
   */
  std::string get_mangled_type_name();

private:

  /**
   * \brief Disabled copy constructor
   */
  Module(const Module &md);
};

} // end of namespace risc

#endif /* MODULE_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
