#include "data_conflict_table.h"

#include <fstream>

#include "mapped_variable.h"
#include "segment.h"
#include "../tools/misc.h"
#include "../tools/variable_helper.h"
#include "../internal_representation/design.h"
#include "../internal_representation/dynamic_internal_representation.h"
#include "../internal_representation/dynamic_path_instance_mapper.h"
#include "../internal_representation/internal_representation_helper.h"
#include "../internal_representation/static_path_instance_mapper.h"

extern DynamicModule *ast_root;

risc::sg::DataConflictTable::DataConflictTable(
    SegmentGraph &graph,
    PathInstanceMapper *path_instance_mapper,
    bool debugging):
  ConflictTable(graph, path_instance_mapper, debugging)
{
  // here we perform a read/write analysis on each segment in the graph
  std::pair<VertexIterator, VertexIterator> vp;
  for(vp = boost::vertices(segment_graph_.graph_);
      vp.first != vp.second; ++vp.first) {

    segment_graph_.graph_[*vp.first].analyze_segment();
  }

  determine_conflict_table();
}

void
risc::sg::DataConflictTable::determine_conflict_table()
{
  bool dynamic_analysis = path_instance_mapper_->perform_dynamic_analysis();
  for(int row=0; row < size_of_conflict_table_; row++) {

    // Here, we are getting the segment id and instance id with the associated
    // row index 'row'.
    std::pair<int, int> row_id_and_instance =
      index_to_segment_and_instance_id(row);

    int row_seg = row_id_and_instance.first;
    int row_inst = row_id_and_instance.second;

    for(int column=0; column <= row; column++) {

      // Here, we are getting the segment id and instance id with the associated
      // column index 'column'.
      std::pair<int, int> column_id_and_instance =
        index_to_segment_and_instance_id(column);

      int column_seg = column_id_and_instance.first;
      int column_inst = column_id_and_instance.second;

      if(column_seg == row_seg && row_inst == column_inst) {

        // If only read access in this segment, there will be no conflict
        if(segment_graph_.graph_[segment_id_to_vertex_descriptor(column_seg)].
            write_variables_.empty()) {

          conflict_table_[row][column] = false;
        } else {

          bool has_conflict = false;

          for(std::set<SgVariableSymbol*>::iterator
              write_iter  = segment_graph_.
                              graph_[segment_id_to_vertex_descriptor(row_seg)].
                                write_variables_.begin();
              write_iter != segment_graph_.
                              graph_[segment_id_to_vertex_descriptor(row_seg)].
                                write_variables_.end();
              write_iter++) {

            MappedVariableList mapped_variables;
            if(dynamic_analysis) {
              mapped_variables
                = get_mapped_variable_dynamically(*write_iter, row_inst);
            } else {
              mapped_variables
                = get_mapped_variable_statically(*write_iter, row_inst);
            }

            for(MappedVariableList::iterator
                mapped_variable  = mapped_variables.begin();
                mapped_variable != mapped_variables.end();
                mapped_variable++) {

              SgVariableSymbol *variable_symbol = (*mapped_variable).symbol_;

              if((tools::is_member_variable(variable_symbol)) ||

                  (tools::is_local_variable(variable_symbol) &&
                  tools::is_static_variable(variable_symbol)) ||

                  tools::is_global_variable(variable_symbol) ||
                  tools::is_namespace_variable(variable_symbol)) {
                has_conflict = true;

                if(debugging_) {

                  // We are displaying the mapped variable instead of the
                  // regular variable ... just in case of a reference
                  Conflict conflict(*mapped_variable);
                  conflict_info_[column][row].conflict_variables_.
                    insert(conflict);
                  conflict_info_[row][column].conflict_variables_.
                    insert(conflict);
                }
              }
            }
          }

          conflict_table_[row][column] = has_conflict;
        }

        if(segment_graph_.graph_[segment_id_to_vertex_descriptor(row_seg)].
            in_conflict_with_all_other_segments){

          conflict_info_[row][column].conflict_variables_ =
            segment_graph_.graph_[segment_id_to_vertex_descriptor(row_seg)].
              conflicts_;
        }

        conflict_table_[row][column] |=
          segment_graph_.graph_[segment_id_to_vertex_descriptor(row_seg)].
              in_conflict_with_all_other_segments;

      } else {

        // Here we perform the conflict analysis of the individual parts
        conflict_table_[row][column] = data_conflict_between(
            segment_id_to_vertex_descriptor(row_seg),
            segment_id_to_vertex_descriptor(column_seg),
            row,
            column,
            row_inst,
            column_inst,
            dynamic_analysis);
        conflict_table_[column][row] = conflict_table_[row][column];
      }
    }
  }
}

void
risc::sg::DataConflictTable::color_members_of_class(
    const MappedVariable &variable,
    DataConflictAstAttributes::Status color)
{
  // we get the type of the variable
  SgInitializedName *var_decl_stmt = variable.symbol_->get_declaration();
  SgType *type = var_decl_stmt->get_type();
  SgDeclarationStatement *type_decl_stmt = type->getAssociatedDeclaration();

  if(!type_decl_stmt) {
    return;
  }

  SgClassDeclaration *class_decl = isSgClassDeclaration(type_decl_stmt);

  if(!class_decl) {
    return;
  }

  class_decl = isSgClassDeclaration(class_decl->get_definingDeclaration());

  SgClassDefinition *class_def = class_decl->get_definition();

  for(std::vector<SgDeclarationStatement*>::iterator
      member_iter  = class_def->get_members().begin();
      member_iter != class_def->get_members().end();
      member_iter++) {

    SgVariableDeclaration *var_decl = isSgVariableDeclaration(*member_iter);
    if(var_decl) {

      if(!var_decl->get_decl_item(0)->get_storageModifier().isStatic()) {


        SgVariableSymbol *var_symbol =
          isSgVariableSymbol(var_decl->get_definition()->get_vardefn()->
                              search_for_symbol_from_symbol_table());

        DataConflictAstAttributes *data_conflict_attribute =
          dynamic_cast<DataConflictAstAttributes*>(
              var_symbol->getAttribute("data_conflict"));

        if(data_conflict_attribute == NULL) {
          data_conflict_attribute = new DataConflictAstAttributes();
          data_conflict_attribute->status_ = color;
          data_conflict_attribute->instance_id_ = variable.instance_id_;
          var_symbol->setAttribute("data_conflict", data_conflict_attribute);

        } else {
          data_conflict_attribute->status_ = color;
          data_conflict_attribute->instance_id_ = variable.instance_id_;
        }
      }
    }
  }
}

void
risc::sg::DataConflictTable::uncolor_members_of_class(
    SgVariableSymbol *var_symbol, int instance_id)
{
  // we get the type of the variable
  SgInitializedName *var_decl_stmt = var_symbol->get_declaration();
  SgType *type = var_decl_stmt->get_type();
  SgDeclarationStatement *type_decl_stmt = type->getAssociatedDeclaration();

  if(!type_decl_stmt) {
    return;
  }

  SgClassDeclaration *class_decl = isSgClassDeclaration(type_decl_stmt);

  if(!class_decl) {
    return;
  }

  class_decl = isSgClassDeclaration(class_decl->get_definingDeclaration());

  SgClassDefinition *class_def = class_decl->get_definition();

  for(std::vector<SgDeclarationStatement*>::iterator
      member_iter  = class_def->get_members().begin();
      member_iter != class_def->get_members().end();
      member_iter++) {

    SgVariableDeclaration *var_decl = isSgVariableDeclaration(*member_iter);
    if(var_decl) {

      if(!var_decl->get_decl_item(0)->get_storageModifier().isStatic()) {

        SgVariableSymbol *var_symbol =
          isSgVariableSymbol(var_decl->get_definition()->get_vardefn()->
                              search_for_symbol_from_symbol_table());

        DataConflictAstAttributes *data_conflict_attribute =
          dynamic_cast<DataConflictAstAttributes*>(
              var_symbol->getAttribute("data_conflict"));

        if(data_conflict_attribute != NULL) {

          data_conflict_attribute->status_ = DataConflictAstAttributes::UNUSED;
          data_conflict_attribute->instance_id_ = instance_id;
        }
      }
    }
  }
}

bool
risc::sg::DataConflictTable::check_members_of_class(
    const MappedVariable &mapped_variable,
    DataConflictAstAttributes::Status color)
{
  SgInitializedName *var_decl_stmt = mapped_variable.symbol_->get_declaration();
  SgType *type = var_decl_stmt->get_type();
  SgDeclarationStatement *type_decl_stmt = type->getAssociatedDeclaration();

  if(!type_decl_stmt) {
    return false;
  }

  // we get the type of the variable
  SgClassDeclaration *class_decl = isSgClassDeclaration(type_decl_stmt);

  if(!class_decl) {
    return false;
  }

  class_decl = isSgClassDeclaration(class_decl->get_definingDeclaration());

  SgClassDefinition *class_def = class_decl->get_definition();

  for(std::vector<SgDeclarationStatement*>::iterator
      member_iter  = class_def->get_members().begin();
      member_iter != class_def->get_members().end();
      member_iter++) {

    SgVariableDeclaration *var_decl = isSgVariableDeclaration(*member_iter);
    if(var_decl) {
      if(!var_decl->get_decl_item(0)->get_storageModifier().isStatic()) {

        DataConflictAstAttributes *data_conflict_attribute =
          dynamic_cast<DataConflictAstAttributes*>(
              mapped_variable.symbol_->getAttribute("data_conflict"));

        if(data_conflict_attribute != NULL) {

          if(mapped_variable.instance_id_
              == data_conflict_attribute->instance_id_) {

            if(data_conflict_attribute->status_ == color) {
              return true;
            }
          }
        }
      }
    }
  }
  return false;
}

bool
risc::sg::DataConflictTable::data_conflict_between(
    VertexDescriptor first,
    VertexDescriptor second,
    int row, int column,
    int inst_id_first, int inst_id_second,
    bool dynamic_analysis)
{
  if(segment_graph_.graph_[second].in_conflict_with_all_other_segments
      || segment_graph_.graph_[first].in_conflict_with_all_other_segments) {

    if(segment_graph_.graph_[first].in_conflict_with_all_other_segments) {

      conflict_info_[row][column].conflict_variables_.insert(
          segment_graph_.graph_[first].conflicts_.begin(),
          segment_graph_.graph_[first].conflicts_.end());

      conflict_info_[column][row].conflict_variables_.insert(
          segment_graph_.graph_[first].conflicts_.begin(),
          segment_graph_.graph_[first].conflicts_.end());
    }

    if(segment_graph_.graph_[second].in_conflict_with_all_other_segments) {

      conflict_info_[row][column].conflict_variables_.insert(
          segment_graph_.graph_[second].conflicts_.begin(),
          segment_graph_.graph_[second].conflicts_.end());

      conflict_info_[column][row].conflict_variables_.insert(
          segment_graph_.graph_[second].conflicts_.begin(),
          segment_graph_.graph_[second].conflicts_.end());
    }

    return true;
  }

  std::set<Conflict> conflicts;

  bool conflict = false;

  for(std::set<SgVariableSymbol*>::iterator
      read_iter  = segment_graph_.graph_[first].read_variables_.begin();
      read_iter != segment_graph_.graph_[first].read_variables_.end();
      read_iter++) {

    MappedVariableList mapped_variables;
    if(dynamic_analysis) {
      mapped_variables
        = get_mapped_variable_dynamically(*read_iter, inst_id_first);
    } else {
      mapped_variables
        = get_mapped_variable_statically(*read_iter, inst_id_first);
    }

    for(MappedVariableList::iterator
        mapped_variable  = mapped_variables.begin();
        mapped_variable != mapped_variables.end();
        mapped_variable++) {

      if((tools::is_local_variable((*mapped_variable).symbol_) &&
          !tools::is_static_variable((*mapped_variable).symbol_)) ||
          tools::is_function_parameter_variable((*mapped_variable).symbol_)) {
        continue;
      }

      DataConflictAstAttributes *data_conflict_attribute =
        dynamic_cast<DataConflictAstAttributes*>(
            ((*mapped_variable).symbol_)->getAttribute("data_conflict"));

      if(data_conflict_attribute == NULL) {
        data_conflict_attribute = new DataConflictAstAttributes();
        data_conflict_attribute->status_ = DataConflictAstAttributes::READ;
        data_conflict_attribute->instance_id_ = (*mapped_variable).instance_id_;
        ((*mapped_variable).symbol_)->setAttribute("data_conflict",
                                       data_conflict_attribute);

      } else {
        data_conflict_attribute->status_ = DataConflictAstAttributes::READ;
        data_conflict_attribute->instance_id_ = (*mapped_variable).instance_id_;
      }

      // In the case that we are reading a non pod data type we have to mark
      // all class/struct members as read
      color_members_of_class(*mapped_variable, DataConflictAstAttributes::READ);
    }
  }

  for(std::set<SgVariableSymbol*>::iterator
      write_iter  = segment_graph_.graph_[first].write_variables_.begin();
      write_iter != segment_graph_.graph_[first].write_variables_.end();
      write_iter++) {

    MappedVariableList mapped_variables;
    if(dynamic_analysis) {
      mapped_variables
        = get_mapped_variable_dynamically(*write_iter, inst_id_first);
    } else {
      mapped_variables
        = get_mapped_variable_statically(*write_iter, inst_id_first);
    }

    for(MappedVariableList::iterator
        mapped_variable  = mapped_variables.begin();
        mapped_variable != mapped_variables.end();
        mapped_variable++) {

      if((tools::is_local_variable((*mapped_variable).symbol_) &&
          !tools::is_static_variable((*mapped_variable).symbol_)) ||
          tools::is_function_parameter_variable((*mapped_variable).symbol_)) {
        continue;
      }

      DataConflictAstAttributes *data_conflict_attribute =
        dynamic_cast<DataConflictAstAttributes*>(
            ((*mapped_variable).symbol_)->getAttribute("data_conflict"));

      if(data_conflict_attribute == NULL) {
        data_conflict_attribute = new DataConflictAstAttributes();
        data_conflict_attribute->status_ = DataConflictAstAttributes::WRITE;
        data_conflict_attribute->instance_id_ = (*mapped_variable).instance_id_;
        ((*mapped_variable).symbol_)->setAttribute("data_conflict",
             data_conflict_attribute);

      } else {
        data_conflict_attribute->status_ = DataConflictAstAttributes::WRITE;
        data_conflict_attribute->instance_id_ = (*mapped_variable).instance_id_;
      }

      // In the case that we are reading a non pod data type we have to mark
      // all class/struct members as write
      color_members_of_class((*mapped_variable),
          DataConflictAstAttributes::WRITE);
    }
  }

  for(std::set<SgVariableSymbol*>::iterator
      read_iter  = segment_graph_.graph_[second].read_variables_.begin();
      read_iter != segment_graph_.graph_[second].read_variables_.end();
      read_iter++) {

    MappedVariableList mapped_variables;
    if(dynamic_analysis) {
      mapped_variables
        = get_mapped_variable_dynamically(*read_iter, inst_id_second);
    } else {
      mapped_variables
        = get_mapped_variable_statically(*read_iter, inst_id_second);
    }

    for(MappedVariableList::iterator
        mapped_variable  = mapped_variables.begin();
        mapped_variable != mapped_variables.end();
        mapped_variable++) {

      if((tools::is_local_variable((*mapped_variable).symbol_) &&
          !tools::is_static_variable((*mapped_variable).symbol_)) ||
          tools::is_function_parameter_variable((*mapped_variable).symbol_)) {
        continue;
      }

      DataConflictAstAttributes *data_conflict_attribute =
        dynamic_cast<DataConflictAstAttributes*>(
            ((*mapped_variable).symbol_)->getAttribute("data_conflict"));

      if(data_conflict_attribute != NULL) {

        if(data_conflict_attribute->status_
            == DataConflictAstAttributes::WRITE) {

          if((*mapped_variable).instance_id_
                 == data_conflict_attribute->instance_id_) {
            conflict = true;
          } else {

            conflict = false;
          }

          if(debugging_) {
            conflicts.insert(*mapped_variable);
          } else {
            break;
          }
        }
      }

      conflict |= check_members_of_class(*mapped_variable,
                      DataConflictAstAttributes::WRITE);
    }
  }

  for(std::set<SgVariableSymbol*>::iterator
      write_iter  = segment_graph_.graph_[second].write_variables_.begin();
      write_iter != segment_graph_.graph_[second].write_variables_.end();
      write_iter++) {

    MappedVariableList mapped_variables;
    if(dynamic_analysis) {
      mapped_variables
        = get_mapped_variable_dynamically(*write_iter, inst_id_second);
    } else {
      mapped_variables
        = get_mapped_variable_statically(*write_iter, inst_id_second);
    }

    for(MappedVariableList::iterator
        mapped_variable  = mapped_variables.begin();
        mapped_variable != mapped_variables.end();
        mapped_variable++) {

      if((tools::is_local_variable((*mapped_variable).symbol_) &&
          !tools::is_static_variable((*mapped_variable).symbol_)) ||
          tools::is_function_parameter_variable((*mapped_variable).symbol_)) {
        continue;
      }

      DataConflictAstAttributes *data_conflict_attribute =
        dynamic_cast<DataConflictAstAttributes*>(
            ((*mapped_variable).symbol_)->getAttribute("data_conflict"));

      if(data_conflict_attribute != NULL) {

        if(data_conflict_attribute->status_) {

          if(data_conflict_attribute->status_ !=
              DataConflictAstAttributes::UNUSED) {

            if((*mapped_variable).instance_id_
                == data_conflict_attribute->instance_id_) {
              conflict = true;
            } else {

              conflict = false;
            }

            if(debugging_) {
              conflicts.insert(*mapped_variable);
            } else {
              break;
            }
          }
        }
      }

      conflict |= check_members_of_class(*mapped_variable,
                      DataConflictAstAttributes::WRITE);
      conflict |= check_members_of_class(*mapped_variable,
                      DataConflictAstAttributes::READ);
    }
  }

  for(std::set<SgVariableSymbol*>::iterator
      read_iter  = segment_graph_.graph_[first].read_variables_.begin();
      read_iter != segment_graph_.graph_[first].read_variables_.end();
      read_iter++) {

    MappedVariableList mapped_variables;
    if(dynamic_analysis) {
      mapped_variables
        = get_mapped_variable_dynamically(*read_iter, inst_id_first);
    } else {
      mapped_variables
        = get_mapped_variable_statically(*read_iter, inst_id_first);
    }

    for(MappedVariableList::iterator
        mapped_variable  = mapped_variables.begin();
        mapped_variable != mapped_variables.end();
        mapped_variable++) {

      if((tools::is_local_variable((*mapped_variable).symbol_) &&

          !tools::is_static_variable((*mapped_variable).symbol_)) ||
          tools::is_function_parameter_variable((*mapped_variable).symbol_)) {
        continue;
      }

      DataConflictAstAttributes *data_conflict_attribute =
        dynamic_cast<DataConflictAstAttributes*>(
            ((*mapped_variable).symbol_)->getAttribute("data_conflict"));
      data_conflict_attribute->status_ = DataConflictAstAttributes::UNUSED;
      data_conflict_attribute->instance_id_ = MappedVariable::InvalidVariable;

      uncolor_members_of_class((*mapped_variable).symbol_,
          MappedVariable::InvalidVariable);
    }
  }

  for(std::set<SgVariableSymbol*>::iterator
      write_iter  = segment_graph_.graph_[first].write_variables_.begin();
      write_iter != segment_graph_.graph_[first].write_variables_.end();
      write_iter++) {

    MappedVariableList mapped_variables;
    if(dynamic_analysis) {
      mapped_variables
        = get_mapped_variable_dynamically(*write_iter, inst_id_first);
    } else {
      mapped_variables
        = get_mapped_variable_statically(*write_iter, inst_id_first);
    }

    for(MappedVariableList::iterator
        mapped_variable  = mapped_variables.begin();
        mapped_variable != mapped_variables.end();
        mapped_variable++) {

      if((tools::is_local_variable((*mapped_variable).symbol_) &&
          !tools::is_static_variable((*mapped_variable).symbol_)) ||
          tools::is_function_parameter_variable((*mapped_variable).symbol_)) {
        continue;
      }

      DataConflictAstAttributes *data_conflict_attribute =
        dynamic_cast<DataConflictAstAttributes*>(
            ((*mapped_variable).symbol_)->getAttribute("data_conflict"));
      data_conflict_attribute->status_ = DataConflictAstAttributes::UNUSED;
      data_conflict_attribute->instance_id_ = MappedVariable::InvalidVariable;

      uncolor_members_of_class((*mapped_variable).symbol_,
          MappedVariable::InvalidVariable);
    }
  }

  if(debugging_) {

    conflict_info_[row][column].conflict_variables_ = conflicts;
    conflict_info_[column][row].conflict_variables_ = conflicts;
  }

  return conflict;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
